package cnpc.ai;

import cnpc.*;
import cnpc.engine.*;

import java.util.*;

public class AI_Controller implements CNPC_AiController
{
	private class PersonalityInfo
	{
		public final String path;
		public final String name;
		public final int weight;

		public PersonalityInfo(String path_in, String name_in, int weight_in)
		{
			path = path_in;
			name = name_in;
			weight = weight_in;
		}
	}

	private final boolean usePersonalityCache = true;

	private final Hashtable<String, Hashtable<String, PersonalityInfo>> personalityCache = new Hashtable<String, Hashtable<String, PersonalityInfo>>();

	private final ArrayList<PersonalityInfo> personalities = new ArrayList<PersonalityInfo>();
	private final Hashtable<String, PersonalityInfo> personalitiesByName = new Hashtable<String, PersonalityInfo>();

	private AI_Personality[] players;

	private String currentVariantPath = null;
		

	public AI_Controller()
	{
	}


	public void setVariant(String variant) throws CNPC_UiKillingException
	{
		try
		{
			currentVariantPath = variant;
			updatePersonalities();
		}
		catch(AI_FatalException e)
		{
			handleFatalException(e);
		}
	}

	public String[] getPersonalitiesForCurrentVariant()
	{
		String[] returnValue = new String[personalities.size()];
		for(int i = 0; i < returnValue.length; i++)
		{
			returnValue[i] = personalities.get(i).name;
		}
		return returnValue;
	}

	public String getWeightedRandomPersonality()
	{
		PersonalityInfo[] personalityArray = getPersonalitiesAsArray();

		ArrayList<PersonalityInfo> personalityList = new ArrayList<PersonalityInfo>();

		for(PersonalityInfo info : personalityArray)
		{
			for(int i = 0; i < info.weight; i++)
			{
				personalityList.add(info);
			}
		}

		Random random = new Random();
		int selection = random.nextInt(personalityList.size());

		PersonalityInfo selectedPersonality = personalityList.get(selection);

		return selectedPersonality.name;
	}

	public String getRandomPersonality()
	{
		String[] personalities = getPersonalitiesForCurrentVariant();
		Random random = new Random();
		int index = random.nextInt(personalities.length);
		return personalities[index];
	}
	





	public void initialize(CN_GameState initialState, String variant, String[] playerPersonalityChoices) throws CNPC_UiKillingException
	{
		try
		{
			setVariant(variant);

			AI_Initializer initializer = new AI_Initializer();
			initializer.setVariant(variant);

			initializer.initializeGame(initialState);

			players = new AI_Personality[playerPersonalityChoices.length];

			for(int i = 0; i < players.length; i++)
			{
				if(playerPersonalityChoices[i] == null)
					continue;

				players[i] = initializer.getPersonality(personalitiesByName.get(playerPersonalityChoices[i]).path);
				if(players[i] == null)
					throw new AI_FatalException("AI_Controller.initialize(): No personality by the name \"" + playerPersonalityChoices[i] + "\" exists for variant \"" + variant + "\"!");
			}
		}
		catch(AI_FatalException e)
		{
			handleFatalException(e);
		}
	}



	public CN_InputMessage processGameState(CN_GameState gameState) throws CNPC_UiKillingException
	{
		AI_Personality currentPersonality = players[gameState.currentPlayer];

		try
		{
			if(currentPersonality == null)
				throw new AI_FatalException("AI_Controller.processGameState: Player is not a computer player!");
			return currentPersonality.getInputMessage(gameState, false);
		}
		catch(AI_FatalException e)
		{
			handleFatalException(e);
			return null;
		}
	}

	public CN_InputMessage tryAgain(CN_GameState gameState) throws CNPC_UiKillingException
	{
		AI_Personality currentPersonality = players[gameState.currentPlayer];

		try
		{
			if(currentPersonality == null)
				throw new AI_FatalException("AI_Controller.processGameState: Player is not a computer player!");
			return currentPersonality.getInputMessage(gameState, true);
		}
		catch(AI_FatalException e)
		{
			handleFatalException(e);
			return null;
		}
	}




	private void updatePersonalities() throws AI_FatalException
	{
		Hashtable<String, PersonalityInfo> cachedPersonalities = personalityCache.get(currentVariantPath);
		if(cachedPersonalities != null)
		{
			updateCachedPersonalities(cachedPersonalities);
			return;
		}

		personalities.clear();
		personalitiesByName.clear();

		AI_Initializer initializer = new AI_Initializer();
		initializer.setVariant(currentVariantPath);

		ArrayList<AI_Initializer.PersonalityInfo> personalityInfo = initializer.getPersonalityInfo();
		PersonalityInfo newInfo;
		for(AI_Initializer.PersonalityInfo info : personalityInfo)
		{
			newInfo = new PersonalityInfo(info.path, info.name, info.weight);
			personalities.add(newInfo);
			personalitiesByName.put(newInfo.name, newInfo);
		}

		if(usePersonalityCache)
			cachePersonalities();
	}

	private void cachePersonalities()
	{
		if(personalityCache.get(currentVariantPath) != null)
			return;

		Hashtable<String, PersonalityInfo> cachedPersonalities = new Hashtable<String, PersonalityInfo>();

		Enumeration<String> keys = personalitiesByName.keys();

		String currentPersonalityName;
		PersonalityInfo currentPersonality;
		while(keys.hasMoreElements())
		{
			currentPersonality = personalitiesByName.get(keys.nextElement());
			cachedPersonalities.put(currentPersonality.name, currentPersonality);
		}

		personalityCache.put(currentVariantPath, cachedPersonalities);
	}

	private void updateCachedPersonalities(Hashtable<String, PersonalityInfo> cachedPersonalities)
	{
		personalities.clear();
		personalitiesByName.clear();

		Enumeration<String> keys = cachedPersonalities.keys();
		ArrayList<String> keysList = new ArrayList<String>();

		while(keys.hasMoreElements())
		{
			keysList.add(keys.nextElement());
		}
		
		keysList = sortList(keysList);

		String currentPersonalityName;
		PersonalityInfo currentPersonality;
		for(int i = 0; i < keysList.size(); i++)
		{
			currentPersonalityName = keysList.get(i);
			currentPersonality = cachedPersonalities.get(currentPersonalityName);
			personalities.add(currentPersonality);
			personalitiesByName.put(currentPersonality.name, currentPersonality);
		}
	}
	
	private ArrayList<String> sortList(ArrayList<String> inputList)
	{
		ArrayList<String> returnValue = new ArrayList<String>();
		
		for(String string : inputList)
		{
			addInSortedOrder(returnValue, string);
		}

		return returnValue;
	}
	
	private void addInSortedOrder(ArrayList<String> inputList, String input)
	{
		int i;
		String currentString;
		for(i = 0; i < inputList.size(); i++)
		{
			currentString = inputList.get(i);
			if(currentString.compareTo(input) > 0)
				break;
		}
		inputList.add(i, input);
	}



	private PersonalityInfo[] getPersonalitiesAsArray()
	{
		PersonalityInfo[] returnValue = new PersonalityInfo[personalities.size()];
		for(int i = 0; i < returnValue.length; i++)
		{
			returnValue[i] = personalities.get(i);
		}
		return returnValue;
	}


	private void handleFatalException(AI_FatalException e) throws CNPC_UiKillingException
	{
		CNPC_ExceptionFrame.showFrame(e);
		throw new CNPC_UiKillingException();
	}


}
