package cnpc.ai;

import java.util.*;
import cnpc.engine.CN_GameState;

class AI_MissionCard extends AI_GameCard
{
	public static enum TargetType {EXACT, MOST_AUTHENTIC, NOT_MOST_AUTHENTIC};

	private final ArrayList<VictoryOption> victoryOptions;

	public AI_MissionCard(String identity_in, String cardName_in, ArrayList<ArrayList<String>> victoryOptionTargets_in, ArrayList<ArrayList<TargetType>> victoryOptionTargetTypes_in, ArrayList<ArrayList<Integer>> victoryOptionRequiredAmounts_in) throws AI_FatalException
	{
		super(identity_in, cardName_in);

		victoryOptions = new ArrayList<VictoryOption>();

		if(victoryOptionTargets_in.size() != victoryOptionTargetTypes_in.size())
		{
			throw new AI_FatalException("AI_MissionCard, constructor: input ArrayLists must have same number of items!");
		}
		if(victoryOptionTargets_in.size() != victoryOptionRequiredAmounts_in.size())
		{
			throw new AI_FatalException("AI_MissionCard, constructor: input ArrayLists must have same number of items!");
		}

		for(int i = 0; i < victoryOptionTargets_in.size(); i++)
		{
			victoryOptions.add(new VictoryOption(victoryOptionTargets_in.get(i), victoryOptionTargetTypes_in.get(i), victoryOptionRequiredAmounts_in.get(i)));
		}
	}


	public boolean isArtifactForMission(AI_Artifact artifact, CN_GameState gameState, ArrayList<AI_Artifact> context) throws AI_FatalException
	{
		for(VictoryOption option : victoryOptions)
		{
			if(option.artifactSatisfies(artifact, gameState, context))
				return true;
		}
		return false;
	}

//	public boolean isArtifactForMission(AI_Artifact artifact) throws AI_FatalException
//	{
//		return isArtifactForMission(artifact, null, null);
//	}




	private class VictoryOption
	{
		public final ArrayList<PrereqTriple> prereqs;

		public VictoryOption(ArrayList<String> targets_in, ArrayList<TargetType> targetTypes_in, ArrayList<Integer> requiredAmounts_in) throws AI_FatalException
		{
			if(targets_in.size() != targetTypes_in.size())
			{
				throw new AI_FatalException("AI_MissionCard.VictoryOption, constructor: input ArrayLists must have same number of items!");
			}
			if(targets_in.size() != requiredAmounts_in.size())
			{
				throw new AI_FatalException("AI_MissionCard.VictoryOption, constructor: input ArrayLists must have same number of items!");
			}

			prereqs = new ArrayList<PrereqTriple>();

			for(int i = 0; i < targets_in.size(); i++)
			{
				prereqs.add(new PrereqTriple(targets_in.get(i), targetTypes_in.get(i), requiredAmounts_in.get(i)));
			}
		}

		public boolean artifactSatisfies(AI_Artifact artifact, CN_GameState gameState, ArrayList<AI_Artifact> context) throws AI_FatalException
		{
			for(PrereqTriple prereq : prereqs)
			{
				if(prereq.artifactSatisfies(artifact, gameState, context))
					return true;
			}
			return false;
		}

	}


	private class PrereqTriple
	{
		public final String target;
		public final TargetType targetType;
		public final int requiredAmount;

		public PrereqTriple(String target_in, TargetType targetType_in, int requiredAmount_in)
		{
			target = target_in;
			targetType = targetType_in;
			requiredAmount = requiredAmount_in;
		}

		//todo: handle other target types
//should be two things: more authentic than anything in play
//and more authentic than anything in context
		public boolean artifactSatisfies(AI_Artifact artifact, CN_GameState gameState, ArrayList<AI_Artifact> context) throws AI_FatalException
		{
			switch(targetType)
			{
				case EXACT:
					if(target.equals(artifact.identity))
						return true;
					break;
				case MOST_AUTHENTIC:
					break;
				case NOT_MOST_AUTHENTIC:
					break;
				default:
					throw new AI_FatalException("AI_MissionCard.PrereqTriple.artifactSatisfies(): Invalid target type!");
			}
			return false;
		}
		
		
		private boolean satisfiesMostAuthentic(AI_Artifact artifact, CN_GameState gameState, ArrayList<AI_Artifact> context)
		{
			if((target == null)||(artifact.authClass == null))
				return false;
			if(!artifact.authClass.equals(target))
				return false;
			
			AI_Artifact mostAuthentic;
			
			mostAuthentic = mostAuthenticInPlay(gameState, target);
			if(mostAuthentic != null)
			{
				if(artifact.authenticity < mostAuthentic.authenticity)
				{
					return false;
				}
			}
			
			mostAuthentic = mostAuthenticInContext(context, target);
			if(mostAuthentic != null)
			{
				if(artifact.authenticity < mostAuthentic.authenticity)
				{
					return false;
				}
			}
			
			return true;
		}
		
		
		private AI_Artifact mostAuthenticInPlay(CN_GameState gameState, String authClass)
		{
			if(gameState == null)
				return null;
			
			ArrayList<AI_Artifact> inPlay = new ArrayList<AI_Artifact>();
			
			AI_Artifact currentArtifact;
			for(CN_GameState.PlayerData player : gameState.players)
			{
				for(Integer cardNumber : player.artifacts)
				{
					currentArtifact = (AI_Artifact)AI_Constants.playCardDirectory.get(cardNumber);
					if(currentArtifact.authClass == null)
						continue;
					if(currentArtifact.authClass.equals(authClass))
						inPlay.add(currentArtifact);
				}
			}
			
			return mostAuthenticInContext(inPlay, authClass);
		}
		
		
		private AI_Artifact mostAuthenticInContext(ArrayList<AI_Artifact> context, String authClass)
		{
			if(context == null)
				return null;
			
			if(context.size() == 0)
				return null;
			
			AI_Artifact highest = context.get(0);
			for(AI_Artifact artifact : context)
			{
				if(artifact.authenticity > highest.authenticity)
					highest = artifact;
			}
			
			return highest;			
		}
		
	}


}
