package cnpc.engine;

import java.util.*;

class CN_IdCard extends CN_GameCard
{
	private final ArrayList<VictoryOption> victoryOptions;

	public CN_IdCard(int cardNumber_in, String identity_in, ArrayList<ArrayList<String>> victoryOptions_in) throws CN_FatalException
	{
		super(cardNumber_in, identity_in, "");
		victoryOptions = new ArrayList<VictoryOption>();

		for(int i = 0; i < victoryOptions_in.size(); i++)
		{
			victoryOptions.add(new VictoryOption(victoryOptions_in.get(i)));
		}
	}



	public boolean testForVictory(CN_Timeline timeline) throws CN_FatalException
	{
		ArrayList<ArrayList<String>> victoryConditionTargetArrays = getVictoryConditionTargetArrays();
		ArrayList<ArrayList<String>> victoryConditionPatchedByArrays = getVictoryConditionPatchedByArrays();

		if(victoryConditionTargetArrays.size() != victoryConditionPatchedByArrays.size())
		{
			throw new CN_FatalException("CN_IdCard, testForVictory(): Target Arrays and Patched By Arrays are not the same size!");
		}

		for(int i = 0; i < victoryConditionTargetArrays.size(); i++)
		{
			if(timeline.isAVictory(victoryConditionTargetArrays.get(i), victoryConditionPatchedByArrays.get(i)))
			{
				return true;
			}
		}

		return false;
	}



	//The following two functions return values that get sent to CN_Timeline's isAVictory function
	private ArrayList<ArrayList<String>> getVictoryConditionTargetArrays()
	{
		ArrayList<ArrayList<String>> returnValue = new ArrayList<ArrayList<String>>();

		VictoryOption option;
		for(int i = 0; i < victoryOptions.size(); i++)
		{
			option = victoryOptions.get(i);
			returnValue.add(option.getVictoryConditionTargets());
		}

		return returnValue;
	}

	private ArrayList<ArrayList<String>> getVictoryConditionPatchedByArrays()
	{
		ArrayList<ArrayList<String>> returnValue = new ArrayList<ArrayList<String>>();

		VictoryOption option;
		for(int i = 0; i < victoryOptions.size(); i++)
		{
			option = victoryOptions.get(i);
			returnValue.add(option.getVictoryConditionPatchedBys());
		}

		return returnValue;
	}



	private class VictoryOption
	{
		public final ArrayList<PrereqPair> prereqs;

		/**
		 * Constructs a new VictoryOption
		 *
		 * @param prereqs_in			An ArrayList of strings and/or nulls.  The number of strings must be even, and every two strings is a pair.  The first string represents a ripplepoint; the second a patch.  If the specified patch is on the specified ripplepoint, then that victory condition is met.  If the patch is null, then the ripplepoint must be unflipped for the victory condition to be met.
		 * @throws CN_FatalException	If the input ArrayList does not have an even number of members, or a ripplepoint string is null
		 */
		public VictoryOption(ArrayList<String> prereqs_in) throws CN_FatalException
		{
			if(prereqs_in.size() % 2 != 0)
			{
				throw new CN_FatalException("CN_IdCard.VictoryOption, constructor: input ArrayList must have an even number of items!");
			}

			prereqs = new ArrayList<PrereqPair>();

			String currentPrereqString;
			String currentTarget = null;
			String currentPatchedBy = null;

			for(int i = 0; i < prereqs_in.size(); i++)
			{
				currentPrereqString = prereqs_in.get(i);

				if(i % 2 == 0)
				{
					currentTarget = currentPrereqString;
					if(currentTarget == null)
					{
						throw new CN_FatalException("CN_IdCard.VictoryOption, constructor: Targets are ripplepoints and cannot be null!");
					}

				}
				else
				{
					currentPatchedBy = currentPrereqString;
					prereqs.add(new PrereqPair(currentTarget, currentPatchedBy));
				}
			}
		}


		public ArrayList<String> getVictoryConditionTargets()
		{
			ArrayList<String> returnValue = new ArrayList<String>();

			PrereqPair prereq;
			for(int i = 0; i < prereqs.size(); i++)
			{
				prereq = prereqs.get(i);
				returnValue.add(prereq.target);
			}

			return returnValue;
		}

		public ArrayList<String> getVictoryConditionPatchedBys()
		{
			ArrayList<String> returnValue = new ArrayList<String>();

			PrereqPair prereq;
			for(int i = 0; i < prereqs.size(); i++)
			{
				prereq = prereqs.get(i);
				returnValue.add(prereq.patchedBy);
			}

			return returnValue;
		}




		public void debugPrint()
		{
			for(PrereqPair prereq : prereqs)
			{
				System.out.print("\n");
				prereq.debugPrint();
			}
		}
	}


	private class PrereqPair
	{
		public final String target;
		public final String patchedBy;

		public PrereqPair(String target_in, String patchedBy_in)
		{
			patchedBy = patchedBy_in;
			target = target_in;
		}

		public PrereqPair(String target_in)
		{
			this(target_in, null);
		}


		public void debugPrint()
		{
			System.out.println("Target: " + target);
			if(patchedBy != null)
				System.out.println("Patched By: " + patchedBy);
			else
				System.out.println("Patched By: Null");
		}
	}




	public void debugPrint()
	{
		System.out.print("\n\n");
		System.out.print("\n" + cardNumber);
		System.out.print("\n" + identity);
		for(VictoryOption victoryOption : victoryOptions)
		{
			System.out.print("\n");
			victoryOption.debugPrint();
		}
	}



}
