package cnpc.engine;

import java.util.*;

class CN_MissionCard extends CN_GameCard
{
	public static enum TargetType {EXACT, MOSTAUTHENTIC, NOTMOSTAUTHENTIC};

	private final ArrayList<VictoryOption> victoryOptions;


	public CN_MissionCard(int cardNumber_in, String identity_in, ArrayList<ArrayList<String>> victoryOptionTargets_in, ArrayList<ArrayList<TargetType>> victoryOptionTargetTypes_in, ArrayList<ArrayList<Integer>> victoryOptionRequiredAmounts_in) throws CN_FatalException
	{
		super(cardNumber_in, identity_in, "");
		victoryOptions = new ArrayList<VictoryOption>();

		if(victoryOptionTargets_in.size() != victoryOptionTargetTypes_in.size())
		{
			throw new CN_FatalException("CN_MissionCard, constructor: input ArrayLists must have same number of items!");
		}
		if(victoryOptionTargets_in.size() != victoryOptionRequiredAmounts_in.size())
		{
			throw new CN_FatalException("CN_MissionCard, constructor: input ArrayLists must have same number of items!");
		}

		for(int i = 0; i < victoryOptionTargets_in.size(); i++)
		{
			victoryOptions.add(new VictoryOption(victoryOptionTargets_in.get(i), victoryOptionTargetTypes_in.get(i), victoryOptionRequiredAmounts_in.get(i)));
		}
	}


	public boolean testForVictory(ArrayList<CN_PlayCard> artifacts, ArrayList<Integer> highestAuthsInPlay) throws CN_FatalException
	{
		if(artifacts.size() != highestAuthsInPlay.size())
		{
			throw new CN_FatalException("CN_MissionCard, testForVictory(): input ArrayLists must have same number of items!");
		}

		for(VictoryOption option : victoryOptions)
		{
			if(option.testVictoryOption(artifacts, highestAuthsInPlay))
			{
				return true;
			}
		}
		return false;
	}



	private class VictoryOption
	{
		public final ArrayList<PrereqTriple> prereqs;

		/**
		 * Constructs a new VictoryOption
		 *
		 * @param targets_in			An ArrayList of strings.  If the corresponding targetType of the string is EXACT, then the string represents an artifact that must be present to satisfy the VictoryOption.  If it is MOSTAUTHENTIC, it represents the AuthClass that the player must have the highest Authenticity of in play to satisfy the VictoryOption.  If it is NOTMOSTAUTHENTIC, it represents the AuthClass that the player must have anything but the highest Authenticity of in play to satisfy the VictoryOption.
		 * @param targetTypes_in		An ArrayList of TargetTypes.  See explanation for targets_in.
		 * @param requiredAmounts_in	An ArrayList of integers.  This is the amount of copies of the card represented by the corresponding string and boolean the player must have for the VictoryOption to be satisfied.
		 * @throws CN_FatalException	If all the input ArrayLists don't have the same number of members.
		 */
		public VictoryOption(ArrayList<String> targets_in, ArrayList<TargetType> targetTypes_in, ArrayList<Integer> requiredAmounts_in) throws CN_FatalException
		{
			if(targets_in.size() != targetTypes_in.size())
			{
				throw new CN_FatalException("CN_MissionCard.VictoryOption, constructor: input ArrayLists must have same number of items!");
			}
			if(targets_in.size() != requiredAmounts_in.size())
			{
				throw new CN_FatalException("CN_MissionCard.VictoryOption, constructor: input ArrayLists must have same number of items!");
			}

			prereqs = new ArrayList<PrereqTriple>();

			for(int i = 0; i < targets_in.size(); i++)
			{
				prereqs.add(new PrereqTriple(targets_in.get(i), targetTypes_in.get(i), requiredAmounts_in.get(i)));
			}
		}


		public boolean testVictoryOption(ArrayList<CN_PlayCard> artifacts, ArrayList<Integer> highestAuthsInPlay) throws CN_FatalException
		{
			if(artifacts.size() != highestAuthsInPlay.size())
			{
				throw new CN_FatalException("CN_MissionCard.VictoryOption, testVictoryOption(): input ArrayLists must have same number of items!");
			}

			int numberOfSatisfyingCards;

			CN_PlayCard artifact;
			int highestAuthInPlay;

			for(PrereqTriple prereq : prereqs)
			{
				numberOfSatisfyingCards = 0;
				for(int i = 0; i < artifacts.size(); i++)
				{
					artifact = artifacts.get(i);

					if(artifact.cardType != CN_PlayCard.PlayCardType.ARTIFACT)
					{
						throw new CN_FatalException("CN_MissionCard.VictoryOption, testVictoryOption(): card " + artifact.identity + " is not an artifact!");
					}

					if(prereq.targetType == TargetType.EXACT)
					{
						if(prereq.target.equals(artifact.identity))
						{
							numberOfSatisfyingCards++;
						}
					}
					else if(prereq.targetType == TargetType.MOSTAUTHENTIC)
					{
						if(highestAuthsInPlay.get(i) == null)
						{
							throw new CN_FatalException("CN_MissionCard.VictoryOption, testVictoryOption(): If target type is MostAuthentic, highest auth in play cannot be null!");
						}
						highestAuthInPlay = highestAuthsInPlay.get(i);

						if((prereq.target.equals(artifact.authClass))&&(artifact.authenticity >= highestAuthInPlay))
						{
							numberOfSatisfyingCards++;
						}

					}
					else if(prereq.targetType == TargetType.NOTMOSTAUTHENTIC)
					{
						if(highestAuthsInPlay.get(i) == null)
						{
							throw new CN_FatalException("CN_MissionCard.VictoryOption, testVictoryOption(): If target type is NotMostAuthentic, highest auth in play cannot be null!");
						}
						highestAuthInPlay = highestAuthsInPlay.get(i);

						if((prereq.target.equals(artifact.authClass))&&(artifact.authenticity < highestAuthInPlay))
						{
							numberOfSatisfyingCards++;
						}
					}
					else
					{
						throw new CN_FatalException("CN_MissionCard.VictoryOption, testVictoryOption(): targetType not set for current prereq!");
					}
				}
				if(numberOfSatisfyingCards < prereq.requiredAmount)
				{
					return false;
				}
			}

			return true;
		}


		public void debugPrint()
		{
			for(PrereqTriple prereq : prereqs)
			{
				System.out.print("\n");
				prereq.debugPrint();
			}
		}

	}


	private class PrereqTriple
	{
		public final String target;
		public final TargetType targetType;
		public final int requiredAmount;

		public PrereqTriple(String target_in, TargetType targetType_in, int requiredAmount_in)
		{
			target = target_in;
			targetType = targetType_in;
			requiredAmount = requiredAmount_in;
		}

		public void debugPrint()
		{
			System.out.println("Target: " + target);
			System.out.println("Target Type: " + targetType);
			System.out.println("Required: " + requiredAmount);
		}

	}

	public void debugPrint()
	{
		System.out.print("\n\n");
		System.out.print("\n" + cardNumber);
		System.out.print("\n" + identity);
		for(VictoryOption victoryOption : victoryOptions)
		{
			System.out.print("\n");
			victoryOption.debugPrint();
		}
	}

}
