package cnpc.engine;

import java.util.*;

class CN_Ripplepoint extends CN_TimelineCard
{
	private final ArrayList<PatchingOption> patchingOptions;
	private final ArrayList<ForceParadox> flipStateOverrides;
	public int patchedBy;

	private boolean isUnstable;

	private boolean isLocked;



	private class ForceParadox
	{
		public final CN_Linchpin linchpin;
		public final FlipState linchpinsFlipState;

		public ForceParadox(CN_Linchpin linchpin_in, FlipState linchpinsFlipState_in)
		{
			linchpin = linchpin_in;
			linchpinsFlipState = linchpinsFlipState_in;
		}
	}




	public CN_Ripplepoint(int cardNumber_in, String identity_in, String cardName_in, int year_in)
	{
		super(cardNumber_in, identity_in, cardName_in, year_in);
		patchingOptions = new ArrayList<PatchingOption>();
		flipStateOverrides = new ArrayList<ForceParadox>();
		isLocked = false;
		patchedBy = CN_Constants.NULL_PATCH;
	}

	public void lockPatchingOptions() throws CN_FatalException
	{
		if(isLocked)
			throw new CN_FatalException("CN_Ripplepoint, lockPatchingOptions(): Patching Options have already been locked!");
		isLocked = true;
	}

	public void addPatchingOption(String patchId, CN_Linchpin[] linchpinPrereqs, FlipState[] flipStatePrereqs) throws CN_FatalException
	{
		if(linchpinPrereqs.length != flipStatePrereqs.length)
			throw new CN_FatalException("CN_Ripplepoint, addPatchingOption(): linchpinPrereqs and flipStatePrereqs don't have the same length!");
		if(isLocked)
			throw new CN_FatalException("CN_Ripplepoint, addPatchingOption(): Patching Options have already been locked!");

		PatchingOption option = new PatchingOption();

		option.patchId = patchId;
		for(int i = 0; i < linchpinPrereqs.length; i++)
		{
			option.addPrereq(linchpinPrereqs[i], flipStatePrereqs[i]);
			linchpinPrereqs[i].addLinkup(this);
		}

		patchingOptions.add(option);
	}

	public void addForceParadox(CN_Linchpin linchpin, boolean flipStateIsFlipped) throws CN_FatalException
	{
		if(isLocked)
			throw new CN_FatalException("CN_Ripplepoint, addForceParadox(): Patching Options and Overrides have already been locked!");

		if(flipStateIsFlipped)
		{
			flipStateOverrides.add(new ForceParadox(linchpin, FlipState.FLIPPED));
		}
		else
		{
			flipStateOverrides.add(new ForceParadox(linchpin, FlipState.UNFLIPPED));
		}


	}



	public Integer updateFlipState() throws CN_FatalException
	{
		Integer returnValue = null;

		if(checkIfFlipped())
		{
			flipState = FlipState.FLIPPED;
		}
		else
		{
			flipState = FlipState.UNFLIPPED;
		}

		if(!checkPatchableBy(patchedBy))
		{
			if(!(patchedBy == CN_Constants.NULL_PATCH))
			{
				returnValue = patchedBy;
				patchedBy = CN_Constants.NULL_PATCH;
			}
		}

		updateIsUnstable();

		return returnValue;
	}

	private void updateIsUnstable() //use only RIGHT AFTER flipState updated
	{
		if(checkIfFlipped())
		{
			isUnstable = false;
			return;
		}

		for(PatchingOption option : patchingOptions)
		{
			if(option.checkPossiblyUnstable())
			{
				isUnstable = true;
				return;
			}
		}
		isUnstable = false;
	}


	public boolean isUnstable()
	{
		return isUnstable;
	}


	public boolean isParadoxed()
	{
		if((flipState == FlipState.FLIPPED) && (patchedBy == CN_Constants.NULL_PATCH))
			return true;
		else
			return false;
	}



	public boolean attemptToPatch(int patchId) throws CN_FatalException
	{
		if(checkPatchableBy(patchId))
		{
			patchedBy = patchId;
			return true;
		}
		else
		{
			return false;
		}
	}

	public boolean checkPatchableBy(int patchId) throws CN_FatalException
	{
		if(patchId == CN_Constants.NULL_PATCH)
			return false;

		String patch = CN_Constants.playCardDirectory.getById(patchId).identity;

		for(PatchingOption option : patchingOptions)
		{
			if((option.patchId.equals(patch)) && (option.checkPrereqs() == true))
				return true;
		}
		return false;
	}


	private boolean checkIfFlipped()
	{
		if(checkOverrides())
		{
			return true;
		}

		for(PatchingOption option : patchingOptions)
		{
			if(option.checkPrereqs() == true)
				return true;
		}
		return false;
	}


	private boolean checkOverrides()
	{
		for(ForceParadox forceParadox : flipStateOverrides)
		{
			if(forceParadox.linchpin.flipState == forceParadox.linchpinsFlipState)
			{
				return true;
			}
		}
		return false;
	}



	private class PatchingOption
	{
		public String patchId;
		private ArrayList<Prereq> prereqs;

		public PatchingOption()
		{
			prereqs = new ArrayList<Prereq>();
		}

		public boolean checkPrereqs()
		{
			for(Prereq prereq : prereqs)
			{
				if(prereq.linkup.flipState != prereq.flipState)
				{
					return false;
				}
			}
			return true;
		}

		public boolean checkPossiblyUnstable()
		{
			if(checkPrereqs() == true)
				return false;


			for(Prereq prereq : prereqs)
			{
				if(prereq.linkup.flipState == prereq.flipState)
				{
					if(prereq.flipState == FlipState.FLIPPED)
					{
						return true;
					}
				}
			}
			return false;
		}


		public void addPrereq(CN_Linchpin linkup, FlipState flipState)
		{
			Prereq prereq = new Prereq();
			prereq.linkup = linkup;
			prereq.flipState = flipState;
			prereqs.add(prereq);
		}


		private class Prereq
		{
			CN_Linchpin linkup;
			FlipState flipState;
		}



		public void debugPrint()
		{
			System.out.println("Patch id: " + patchId);
			for(Prereq prereq : prereqs)
			{
				if(prereq.flipState == FlipState.FLIPPED)
					System.out.println(prereq.linkup.identity + " - FLIPPED");
				else
					System.out.println(prereq.linkup.identity + " - UNFLIPPED");
			}
		}

	}




	public void printBasicInfo()
	{
		super.printBasicInfo();
		System.out.println("PATCHING OPTIONS:");
		for(int i = 0; i < patchingOptions.size(); i++)
		{
			patchingOptions.get(i).debugPrint();
		}
	}




}

