package cnpc.engine;

import java.util.*;

class CN_Timeline
{
	private final CN_TimelineCard[] timeline;

	private final ArrayList<CN_PlayCard> terminalPatches;
	private int terminalYear;


	public CN_Timeline(CN_TimelineCard[] timeline_in)
	{
		timeline = timeline_in;
		terminalPatches = new ArrayList<CN_PlayCard>();
		terminalYear = CN_Constants.NON_TERMINAL_YEAR;
	}


	public ArrayList<CN_PlayCard> resetTimeline() throws CN_FatalException
	{
		ArrayList<CN_PlayCard> returnValue;
		ArrayList<Integer> discardedPatches = new ArrayList<Integer>();
		ArrayList<Integer> currentDiscardedPatches = new ArrayList<Integer>();

		CN_Linchpin linchpin;
		for(CN_TimelineCard card : timeline)
		{
			if(card instanceof CN_Linchpin)
			{
				linchpin = (CN_Linchpin)card;
				if(linchpin.flipState == CN_TimelineCard.FlipState.FLIPPED)
				{
					currentDiscardedPatches = linchpin.flip();
					for(int i = 0; i < currentDiscardedPatches.size(); i++)
					{
						discardedPatches.add(currentDiscardedPatches.get(i));
					}
				}
			}
		}

		returnValue = getActualCards(discardedPatches);
		return returnValue;
	}



	public ArrayList<CN_PlayCard> refreshTimeline() throws CN_FatalException
	{
		ArrayList<CN_PlayCard> returnValue;
		ArrayList<Integer> discardedPatches = new ArrayList<Integer>();

		CN_Ripplepoint ripplepoint;
		for(CN_TimelineCard card : timeline)
		{
			if(card instanceof CN_Ripplepoint)
			{
				ripplepoint = (CN_Ripplepoint)card;
				discardedPatches.add(ripplepoint.updateFlipState());
			}
		}

		returnValue = getActualCards(discardedPatches);
		return returnValue;
	}


	private ArrayList<CN_PlayCard> getActualCards(ArrayList<Integer> discardedPatches) throws CN_FatalException
	{
		ArrayList<CN_PlayCard> returnValue = new ArrayList<CN_PlayCard>();
		for(int i = 0; i < discardedPatches.size(); i++)
		{
			if(discardedPatches.get(i) != null)
			{
				returnValue.add(CN_Constants.playCardDirectory.getById(discardedPatches.get(i)));
			}
		}
		return returnValue;
	}


	public int getTerminalYear()
	{
		return terminalYear;
	}


	public int getUnpatchedParadoxes()
	{
		int returnValue = 0;

		CN_Ripplepoint ripplepoint;

		for(CN_TimelineCard card : timeline)
		{
			if(card instanceof CN_Ripplepoint)
			{
				ripplepoint = (CN_Ripplepoint)card;
				if(ripplepoint.isParadoxed())
				{
					returnValue++;
				}
			}
		}

		return returnValue;
	}


	public boolean isBeforeTerminalYear(CN_PlayCard card)
	{
		if(card.year <= terminalYear)
		{
			return true;
		}
		else
		{
			return false;
		}
	}



	private void addTerminalPatch(CN_PlayCard patch) throws CN_FatalException
	{
		if(patch.cardType != CN_PlayCard.PlayCardType.PATCH)
		{
			throw new CN_FatalException("CN_Timeline, setTerminalPatch(): specified card \"" + patch.identity + "\" is not a patch!");
		}
		terminalPatches.add(patch);
		updateTerminalYear();
	}


	private void updateTerminalYear()
	{
		terminalYear = CN_Constants.NON_TERMINAL_YEAR;
		for(CN_PlayCard patch : terminalPatches)
		{
			if(patch.year < terminalYear)
			{
				terminalYear = patch.year;
			}
		}
	}


	private void removeTerminalPatch(int cardNumber) throws CN_FatalException
	{
		CN_PlayCard removedPatch = null;
		for(CN_PlayCard patch : terminalPatches)
		{
			if(patch.cardNumber == cardNumber)
			{
				removedPatch = patch;
			}
		}

		if(removedPatch != null)
		{
			terminalPatches.remove(removedPatch);
		}
		else
		{
			throw new CN_FatalException("CN_Timeline, removeTerminalPatch(): Specified patch did not exist in terminal patches!");
		}

		updateTerminalYear();
	}


	private boolean isInTerminalPatches(int cardNumber)
	{
		for(CN_PlayCard patch : terminalPatches)
		{
			if(patch.cardNumber == cardNumber)
			{
				return true;
			}
		}
		return false;
	}



	public void banAllLinchpins()
	{
		for(CN_TimelineCard linchpin : timeline)
		{
			if(linchpin instanceof CN_Linchpin)
			{
				((CN_Linchpin)linchpin).flipLegal = false;
				((CN_Linchpin)linchpin).unflipLegal = false;
			}
		}
	}


	private void legalizeAllLinchpinsFlip()
	{
		for(CN_TimelineCard linchpin : timeline)
		{
			if(linchpin instanceof CN_Linchpin)
			{
				((CN_Linchpin)linchpin).flipLegal = true;
			}
		}
	}


	private void legalizeAllLinchpinsUnflip()
	{
		for(CN_TimelineCard linchpin : timeline)
		{
			if(linchpin instanceof CN_Linchpin)
			{
				((CN_Linchpin)linchpin).unflipLegal = true;
			}
		}
	}


	private void legalizeAllLinchpins()
	{
		for(CN_TimelineCard linchpin : timeline)
		{
			if(linchpin instanceof CN_Linchpin)
			{
				((CN_Linchpin)linchpin).flipLegal = true;
				((CN_Linchpin)linchpin).unflipLegal = true;
			}
		}
	}



	public boolean isLinchpinLegal(String linchpinString) throws CN_FatalException
	{
		CN_Linchpin linchpin = findLinchpin(linchpinString);

		if(linchpin == null)
		{
			throw new CN_FatalException("CN_Timeline, isLinchpinLegal(): linchpin string sent (" + linchpinString + ") does not match a linchpin!");
		}


		if(linchpin.year > terminalYear)
		{
			return false;
		}

		if(linchpin.flipState == CN_TimelineCard.FlipState.FLIPPED)
		{
			return linchpin.unflipLegal;
		}
		else if(linchpin.flipState == CN_TimelineCard.FlipState.UNFLIPPED)
		{
			return linchpin.flipLegal;
		}
		else
		{
			return false;
		}

	}


	public void parseBascomLegalityArray(ArrayList<String> legalityArray) throws CN_FatalException
	{
		if((legalityArray.size() % 2 != 0)&&(legalityArray.size() != 1))
		{
			throw new CN_FatalException("parseBascomLegalityArray: The legality array is based on pairs, and must have an even number of entries!");
		}

		banAllLinchpins();

		CN_Linchpin currentLinchpin = null;
		String legalityString;

		if(legalityArray.size() == 1)
		{
			legalityString = legalityArray.get(0);
			if(legalityString.toLowerCase().equals(CN_Constants.TIMELINEFLIP))
			{
				legalizeAllLinchpinsFlip();
			}
			else if(legalityString.toLowerCase().equals(CN_Constants.TIMELINEUNFLIP))
			{
				legalizeAllLinchpinsUnflip();
			}
			else if(legalityString.toLowerCase().equals(CN_Constants.TIMELINEBOTH))
			{
				legalizeAllLinchpins();
			}
			else
			{
				throw new CN_FatalException("parseBascomLegalityArray: Legality array had only one entry, and legality string \"" + legalityString + "\" was not recognized!");
			}
		}
		else
		{
			for(int i = 0; i < legalityArray.size(); i++)
			{
				if(i % 2 == 0)
				{
					currentLinchpin = findLinchpin(legalityArray.get(i));
				}
				else
				{
					if(currentLinchpin == null)
					{
						throw new CN_FatalException("parseBascomLegalityArray: currentLinchpin was null; possibly means string was sent that does not represent a linchpin!");
					}

					legalityString = legalityArray.get(i);
					if(legalityString.toLowerCase().equals(CN_Constants.TIMELINEFLIP))
					{
						currentLinchpin.flipLegal = true;
					}
					else if(legalityString.toLowerCase().equals(CN_Constants.TIMELINEUNFLIP))
					{
						currentLinchpin.unflipLegal = true;
					}
					else if(legalityString.toLowerCase().equals(CN_Constants.TIMELINEBOTH))
					{
						currentLinchpin.flipLegal = true;
						currentLinchpin.unflipLegal = true;
					}
					else
					{
						throw new CN_FatalException("parseBascomLegalityArray: Legality string \"" + legalityString + "\" was not recognized!");
					}
				}
			}
		}
	}




	public boolean isAVictory(ArrayList<String> ripplepoints, ArrayList<String> patchedBy) throws CN_FatalException
	{
		boolean returnValue = true;

		if(ripplepoints.size() != patchedBy.size())
		{
			throw new CN_FatalException("CN_Timeline, isAVictory(): input arraylists must be of the same size!");
		}

		CN_Ripplepoint currentRipplepoint;
		String currentPatch;
		for(int i = 0; i < ripplepoints.size(); i++)
		{
			currentRipplepoint = findRipplepoint(ripplepoints.get(i));
			currentPatch = patchedBy.get(i);

			if(currentRipplepoint == null)
			{
				throw new CN_FatalException("CN_Timeline, isAVictory(): ripplepoint " + ripplepoints.get(i) + " not found!");
			}

			if(currentPatch == null)
			{
				if(currentRipplepoint.flipState != CN_TimelineCard.FlipState.UNFLIPPED)
				{
					returnValue = false;
				}
			}
			else
			{
				if(currentRipplepoint.flipState != CN_TimelineCard.FlipState.FLIPPED)
				{
					returnValue = false;
				}

				if(CN_Constants.playCardDirectory.getById(currentRipplepoint.patchedBy) == null)
				{
					returnValue = false;
				}
				else if(!(CN_Constants.playCardDirectory.getById(currentRipplepoint.patchedBy).identity.equals(currentPatch)))
				{
					returnValue = false;
				}
			}
		}

		return returnValue;
	}




	private CN_TimelineCard findTimelineCard(int cardNumber) throws CN_FatalException
	{
		return findTimelineCard(CN_Constants.timelineCardDirectory.getById(cardNumber).identity);
	}

	private CN_TimelineCard findLinchpin(int cardNumber) throws CN_FatalException
	{
		return findLinchpin(CN_Constants.timelineCardDirectory.getById(cardNumber).identity);
	}

	private CN_TimelineCard findRipplepoint(int cardNumber) throws CN_FatalException
	{
		return findRipplepoint(CN_Constants.timelineCardDirectory.getById(cardNumber).identity);
	}


	private CN_TimelineCard findTimelineCard(String cardIdentity)
	{
		for(CN_TimelineCard card : timeline)
		{
			if(card.identity.equals(cardIdentity))
			{
				return card;
			}
		}
		return null;
	}


	private CN_Linchpin findLinchpin(String linchpinIdentity)
	{
		CN_TimelineCard card = findTimelineCard(linchpinIdentity);
		if(card == null) return null;

		if(card instanceof CN_Linchpin)
		{
			return (CN_Linchpin)(card);
		}
		else
		{
			return null;
		}
	}


	private CN_Ripplepoint findRipplepoint(String ripplepointIdentity)
	{
		CN_TimelineCard card = findTimelineCard(ripplepointIdentity);
		if(card == null) return null;

		if(card instanceof CN_Ripplepoint)
		{
			return (CN_Ripplepoint)(card);
		}
		else
		{
			return null;
		}
	}




	public int timesPatchPresent(CN_PlayCard patch)
	{
		int patchCount = 0;

		for(CN_TimelineCard card : timeline)
		{
			if(card instanceof CN_Ripplepoint)
			{
				if(((CN_Ripplepoint)(card)).patchedBy == patch.cardNumber)
				{
					patchCount++;
				}
			}
		}

		return patchCount;
	}




	public void applyPatch(CN_PlayCard patch, String ripplepointString, boolean isTerminal) throws CN_FatalException
	{
		CN_Ripplepoint ripplepoint = findRipplepoint(ripplepointString);
		if(ripplepoint == null)
		{
			throw new CN_FatalException("CN_Timeline, applyPatch(): Ripplepoint \"" + ripplepointString + "\" does not exist!");
		}

		if(!ripplepoint.attemptToPatch(patch.cardNumber))
		{
			throw new CN_FatalException("CN_Timeline, applyPatch(): Patch \"" + patch.identity + "\" could not be applied to target ripplepoint \"" + ripplepointString + "\"!");
		}

		if(ripplepoint.year > terminalYear)
		{
			throw new CN_FatalException("CN_Timeline, applyPatch(): Attempt to apply a patch after the Terminal Year!");
		}

		if(isTerminal)
		{
			addTerminalPatch(patch);
		}
	}


	public boolean isPatchLegal(CN_PlayCard patch, String ripplepointString) throws CN_FatalException
	{
		if(patch == null)
		{
			throw new CN_FatalException("CN_Timeline, isPatchLegal(): Patch sent was null!");
		}

		CN_Ripplepoint ripplepoint = findRipplepoint(ripplepointString);
		if(ripplepoint == null)
		{
			throw new CN_FatalException("CN_Timeline, isPatchLegal(): Ripplepoint \"" + ripplepointString + "\" does not exist!");
		}

		if(ripplepoint.year > terminalYear)
		{
			return false;
		}

		return ripplepoint.checkPatchableBy(patch.cardNumber);
	}


	public boolean isPatchLegalAnywhere(CN_PlayCard patch) throws CN_FatalException
	{
		if(patch == null)
		{
			throw new CN_FatalException("CN_Timeline, isPatchLegal(): Patch sent was null!");
		}

		for(CN_TimelineCard card : timeline)
		{
			if(findRipplepoint(card.identity) != null)
			{
				if(isPatchLegal(patch, card.identity))
				{
					return true;
				}
			}
		}
		return false;
	}



	public boolean isLinchpinFlipped(String linchpinString) throws CN_FatalException
	{
		CN_Linchpin linchpin = findLinchpin(linchpinString);
		if(linchpin != null)
		{
			return linchpin.getIsFlipped();
		}
		else
		{
			throw new CN_FatalException("CN_Timeline, isLinchpinFlipped(): linchpin \"" + linchpinString + "\" is not a valid linchpin!");
		}
	}


	public boolean testAllLinchpins(boolean forFlipped)
	{
		int flips = 0;
		int unflips = 0;
		for(CN_TimelineCard card : timeline)
		{
			if(findLinchpin(card.identity) != null)
			{
				if(card.getIsFlipped())
				{
					flips++;
				}
				else
				{
					unflips++;
				}
			}
		}

		if(flips == 0)
		{
			return !forFlipped;
		}
		else if(unflips == 0)
		{
			return forFlipped;
		}
		else
		{
			return false;
		}

	}




	public ArrayList<Integer> flipLinchpin(String linchpinString) throws CN_FatalException
	{
		ArrayList<Integer> returnValue;

		CN_Linchpin linchpin = findLinchpin(linchpinString);
		if(linchpin != null)
		{
			if(linchpin.year > terminalYear)
			{
				throw new CN_FatalException("CN_Timeline, flipLinchpin(): Attempt to flip a linchpin after the Terminal Year!");
			}


			returnValue = linchpin.flip();
			for(Integer cardNumber : returnValue)
			{
				if(cardNumber != null)
				{
					if(isInTerminalPatches(cardNumber))
					{
						removeTerminalPatch(cardNumber);
					}
				}
			}
			return returnValue;
		}
		else
		{
			return null;
		}
	}



	public CN_GameState.TimelineCardData[] toGameStateTimeline(CN_PlayCard patchInQuestion) throws CN_FatalException
	{
		CN_GameState.TimelineCardData[] returnValue;
		returnValue = new CN_GameState.TimelineCardData[timeline.length];

		for(int i = 0; i < timeline.length; i++)
		{
			if(timeline[i] instanceof CN_Linchpin)
			{
				returnValue[i] = new CN_GameState.TimelineCardData(timeline[i].cardNumber, timeline[i].getIsFlipped(), timeline[i].getPatchedByForGs(), false, isLinchpinLegal(timeline[i].identity), false, timeline[i].year);
			}
			else
			{
				if(patchInQuestion != null)
				{
					returnValue[i] = new CN_GameState.TimelineCardData(timeline[i].cardNumber, timeline[i].getIsFlipped(), timeline[i].getPatchedByForGs(), isTerminal(timeline[i].getPatchedByForGs()), isPatchLegal(patchInQuestion, timeline[i].identity), isRipplepointUnstable(timeline[i].identity), timeline[i].year);
				}
				else
				{
					returnValue[i] = new CN_GameState.TimelineCardData(timeline[i].cardNumber, timeline[i].getIsFlipped(), timeline[i].getPatchedByForGs(), false, false, isRipplepointUnstable(timeline[i].identity), timeline[i].year);
				}
			}
		}

		return returnValue;
	}



	private boolean isRipplepointUnstable(String ripplepointString) throws CN_FatalException
	{
		CN_Ripplepoint ripplepoint = findRipplepoint(ripplepointString);

		if(ripplepoint == null)
		{
			throw new CN_FatalException("CN_Timeline, isRipplepointUnstable(): ripplepoint string sent (" + ripplepointString + ") does not match a ripplepoint!");
		}
		
		if(ripplepoint.isUnstable())
		{
			return true;
		}
		else
		{
			return false;
		}
	}



	private boolean isTerminal(int patch)
	{
		for(CN_PlayCard currentCard : terminalPatches)
		{
			if(currentCard.cardNumber == patch)
			{
				return true;
			}
		}
		return false;
	}



	public void fromGameStateTimeline(CN_GameState.TimelineCardData[] timelineCardData) throws CN_FatalException
	{
		CN_TimelineCard currentCard;
		CN_Ripplepoint currentRipplepoint;
		for(CN_GameState.TimelineCardData cardData : timelineCardData)
		{
			currentCard = findTimelineCard(cardData.cardNumber);
			if(cardData.isFlipped)
			{
				currentCard.flipState = CN_TimelineCard.FlipState.FLIPPED;
			}
			else
			{
				currentCard.flipState = CN_TimelineCard.FlipState.UNFLIPPED;
			}
		}


		for(CN_GameState.TimelineCardData cardData : timelineCardData)
		{
			currentCard = findTimelineCard(cardData.cardNumber);
			if(currentCard instanceof CN_Ripplepoint)
			{
				currentRipplepoint = (CN_Ripplepoint)currentCard;
				if(cardData.patchedBy != CN_Constants.NULL_PATCH)
				{
					applyPatch(CN_Constants.playCardDirectory.getById(cardData.patchedBy), currentRipplepoint.identity, cardData.isTerminal);
				}
			}
		}

	}





	public void printBasicInfo()
	{
		for(CN_TimelineCard card : timeline)
		{
			card.printBasicInfo();
			System.out.println("\n");
		}
	}
}
