package cnpc.ui;

import cnpc.engine.*;

import javax.swing.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

class UI_MainFrame extends JFrame
{
	private enum GameCardType {PLAYCARD, TIMELINECARD, IDCARD, MISSIONCARD};

	private final SpringLayout layout = new SpringLayout();
	private final Container content = getContentPane();

	private final JPanel mainPanel = new JPanel();
	private final SpringLayout panelLayout = new SpringLayout();

	private final JPanel messagePanel = new JPanel();
	private final SpringLayout messagePanelLayout = new SpringLayout();
	private final JLabel messageLabel = new JLabel("Status");

	private final UI_PlayerPane[] playerPanes;
	private final JLabel[] nameLabels;
	private final String[] playerNames;

	private final JLabel[] personalityNameLabels;
	private final String[] personalityNames;

	private final UI_TimelineFrame timeline;

	private final UI_Messenger messenger;

	private final ArrayList<JComponent> disposableComponents = new ArrayList<JComponent>();

	private final JButton toggleTimeline = new JButton("Show Timeline");;
	private final JButton nextTurn = new JButton("Begin Turn");
	private final JButton discardOne = new JButton("Discard One");
	private final JButton discardTwo = new JButton("Discard Two");
	private final JButton cancel = new JButton("Cancel");


	private boolean timelineFrameJustShown;

	private boolean disposed = false;



	private UI_SearchFrame currentSearchFrame = null;
	private UI_SelectorFrame currentSelectorFrame = null;


	public UI_MainFrame(UI_Messenger messenger_in, CN_GameState initialState, String variant, String[] playerNames_in, String[] personalityNames_in) throws UI_FatalException
	{
		messenger = messenger_in;
		playerNames = playerNames_in;
		personalityNames = personalityNames_in;

		if(initialState.players.length != playerNames.length)
			throw new UI_FatalException("UI_MainFrame, constructor: number of player names does not equal number of players!");

		if(initialState.players.length != personalityNames_in.length)
			throw new UI_FatalException("UI_MainFrame, constructor: number of player personality names does not equal number of players!");

		content.setLayout(layout);

		setTitle(UI_Constants.GAME_NAME);
		setResizable(false);

		initMenu();

		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		initGlobals(variant);

		UI_Initializer initializer = new UI_Initializer();
		initializer.initializeUI(initialState, variant);

		timeline = new UI_TimelineFrame();

		playerPanes = new UI_PlayerPane[initialState.players.length];
		nameLabels = new JLabel[initialState.players.length];
		personalityNameLabels = new JLabel[initialState.players.length];

 		initMainPanel();
		initMessagePanel();
		setProperSize();

		createPlayers(initialState.players);
		createButtons(initialState.currentInputState);

		renderGameState(initialState, false);

		setVisible(true);
	}

	private void initGlobals(String variant)
	{
		String blankImagePath = UI_Constants.PATH_TO_DATA + variant + UI_Constants.BLANK_IMAGE_PATH;
		String unstableImagePath = UI_Constants.PATH_TO_DATA + variant + UI_Constants.UNSTABLE_IMAGE_PATH;
		String xImagePath = UI_Constants.PATH_TO_DATA + variant + UI_Constants.X_IMAGE_PATH;
		String endOfUniverseImagePath = UI_Constants.PATH_TO_DATA + variant + UI_Constants.END_OF_UNIVERSE_IMAGE_PATH;
		UI_Globals.blankImage = new ImageIcon(blankImagePath);
		UI_Globals.xImage = new ImageIcon(xImagePath);
		UI_Globals.unstableImage = new ImageIcon(unstableImagePath);
		UI_Globals.endOfUniverseImage = new ImageIcon(endOfUniverseImagePath);
	}

	private void initMenu()
	{
		final UI_MainFrame thisFrame = this;

		JMenuBar menuBar = new JMenuBar();
		JMenu gameMenu = new JMenu("Game");
		JMenu helpMenu = new JMenu("Help");
		JMenuItem newGameItem = new JMenuItem("New Game");
		JMenuItem exitItem = new JMenuItem("Exit");
		JMenuItem helpItem = new JMenuItem("Chrononauts PC Help");
		JMenuItem aboutItem = new JMenuItem("About");

		gameMenu.add(newGameItem);
		gameMenu.add(new JSeparator());
		gameMenu.add(exitItem);

		helpMenu.add(helpItem);
		helpMenu.add(aboutItem);

		menuBar.add(gameMenu);
		menuBar.add(helpMenu);

		setJMenuBar(menuBar);

		newGameItem.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					messenger.setNewGame();
				}
			}
		);

		exitItem.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					System.exit(0);
				}
			}
		);


		helpItem.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					File fileToOpen = new File("../help/index.html");
					String errorMessage;

					try
					{
						errorMessage = cnpc.CNPC_BrowserLauncher.openURL(fileToOpen.getCanonicalPath());
					}
					catch(IOException f)
					{
						errorMessage = "Unable to open browser because canonical pathname could not be resolved!";
					}

					if(errorMessage != null)
					{
						JOptionPane.showMessageDialog(thisFrame, errorMessage, "Error", JOptionPane.ERROR_MESSAGE);
 					}
				}
			}
		);

		aboutItem.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					new UI_AboutDialog(thisFrame);
				}
			}
		);

	}


	private void initMainPanel()
	{
		mainPanel.setLayout(panelLayout);

		mainPanel.setPreferredSize(new Dimension(UI_Constants.MAIN_PANEL_WIDTH, UI_Constants.MAIN_PANEL_HEIGHT));
		layout.putConstraint(SpringLayout.WEST, mainPanel, UI_Constants.MAIN_FRAME_WEST_BUFFER, SpringLayout.WEST, content);
		layout.putConstraint(SpringLayout.NORTH, mainPanel, UI_Constants.MAIN_FRAME_NORTH_BUFFER, SpringLayout.NORTH, content);

		mainPanel.setBorder(BorderFactory.createLoweredBevelBorder());

		content.add(mainPanel);
	}

	private void initMessagePanel()
	{
		messageLabel.setFont(UI_Constants.STATUS_FONT);

		messagePanel.setLayout(messagePanelLayout);
		messagePanel.setPreferredSize(new Dimension(UI_Constants.MESSAGE_PANEL_WIDTH, UI_Constants.MESSAGE_PANEL_HEIGHT));
		layout.putConstraint(SpringLayout.WEST, messagePanel, UI_Constants.MAIN_FRAME_WEST_BUFFER, SpringLayout.WEST, content);
		layout.putConstraint(SpringLayout.NORTH, messagePanel, UI_Constants.MAIN_MESSAGE_PANEL_SEPARATION, SpringLayout.SOUTH, mainPanel);
		messagePanel.setBorder(BorderFactory.createLoweredBevelBorder());
		content.add(messagePanel);

		int messageLabelOffset = (messagePanel.getPreferredSize().height - messageLabel.getPreferredSize().height) / 2 - 1;
		messagePanelLayout.putConstraint(SpringLayout.WEST, messageLabel, 10, SpringLayout.WEST, messagePanel);
		messagePanelLayout.putConstraint(SpringLayout.NORTH, messageLabel, messageLabelOffset, SpringLayout.NORTH, messagePanel);
		messagePanel.add(messageLabel);
	}

	private void setProperSize()
	{
		int width = UI_Constants.MAIN_FRAME_WEST_BUFFER + mainPanel.getPreferredSize().width + UI_Constants.MAIN_FRAME_EAST_BUFFER;
		int height = UI_Constants.TITLE_BAR_HEIGHT + UI_Constants.MENU_BAR_HEIGHT + UI_Constants.MAIN_FRAME_NORTH_BUFFER + mainPanel.getPreferredSize().height + UI_Constants.MAIN_MESSAGE_PANEL_SEPARATION + messagePanel.getPreferredSize().height + UI_Constants.MAIN_FRAME_SOUTH_BUFFER;
		setSize(width, height);
	}








	private void renderForEndGame(final CN_GameState gameState) throws UI_FatalException
	{
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		addWindowListener
		(
			new WindowAdapter()
			{
				public void windowClosing(WindowEvent e)
				{
					performDisposal();
				}
			}
		);
		setPersonalityNameLabelsVisible(true);

		timelineFrameJustShown = false;

		if(gameState == null)
		{
			throw new UI_FatalException("UI_MainFrame.renderForEndGame(): Game state was null!");
		}

		if(currentSearchFrame != null)
		{
			currentSearchFrame.dispose();
			currentSearchFrame = null;
		}
		if(currentSelectorFrame != null)
		{
			currentSelectorFrame.dispose();
			currentSelectorFrame = null;
		}

		while(!disposableComponents.isEmpty())
		{
			mainPanel.remove(disposableComponents.get(0));
			disposableComponents.remove(0);
		}

		updateButtonsForEndGame();

		renderPlayers(gameState.players, gameState.currentPlayer, true);

		if(!gameState.deck.empty())
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(getGameCard(gameState.deck.peek(), GameCardType.PLAYCARD), true), UI_Constants.DECKPOSX, UI_Constants.DECKPOSY));
		}
		else
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(), UI_Constants.DECKPOSX, UI_Constants.DECKPOSY));
		}

		if(!gameState.discards.empty())
		{
			final UI_CardPanel currentPanel = new UI_CardPanel(getGameCard(gameState.discards.peek(), GameCardType.PLAYCARD), false);
			final Stack<UI_CardPanel> searchFrameStack = convertToCardPanelStack(gameState.discards, GameCardType.PLAYCARD);
			currentPanel.addMouseListener
			(
				new MouseAdapter()
				{
					public void mouseClicked(MouseEvent e)
					{
						if((e.getButton() == MouseEvent.BUTTON1)&&(e.getClickCount() % 2 == 0))
						{
							new UI_SearchFrame("Discard Pile", searchFrameStack);
						}
					}
				}
			);

			disposableComponents.add(placeMainPanelComponent(currentPanel, UI_Constants.DISCARDSPOSX, UI_Constants.DISCARDSPOSY));
		}
		else
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(), UI_Constants.DISCARDSPOSX, UI_Constants.DISCARDSPOSY));
		}

		renderTimeline(gameState.timeline, gameState.currentInputState, gameState.terminalYear);

		JLabel unpatchedParadoxLabel = new JLabel("Unpatched Paradoxes: " + String.valueOf(gameState.unpatchedParadoxes));
		unpatchedParadoxLabel.setFont(UI_Constants.UNPATCHED_PARADOXES_FONT);
		if(gameState.unpatchedParadoxes > UI_Constants.UNPATCHED_PARADOX_LABEL_HIGH_THRESHOLD)
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_HIGH_COLOR);
		else if(gameState.unpatchedParadoxes > UI_Constants.UNPATCHED_PARADOX_LABEL_MID_THRESHOLD)
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_MID_COLOR);
		else
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_LOW_COLOR);

		disposableComponents.add(placeMainPanelComponent(unpatchedParadoxLabel, UI_Constants.UNPATCHED_PARADOXES_X, UI_Constants.UNPATCHED_PARADOXES_Y));

		setTimelineVisible(false);

		updateStatus("The game is over.");

		validate();
	}



	private void renderForComputer(CN_GameState gameState) throws UI_FatalException
	{
		if(currentSearchFrame != null)
		{
			currentSearchFrame.dispose();
			currentSearchFrame = null;
		}
		if(currentSelectorFrame != null)
		{
			currentSelectorFrame.dispose();
			currentSelectorFrame = null;
		}

		while(!disposableComponents.isEmpty())
		{
			mainPanel.remove(disposableComponents.get(0));
			disposableComponents.remove(0);
		}

		updateButtonsForComputer();

		renderPlayers(gameState.players, UI_Constants.NULLPLAYER);

		if(!gameState.deck.empty())
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(getGameCard(gameState.deck.peek(), GameCardType.PLAYCARD), true), UI_Constants.DECKPOSX, UI_Constants.DECKPOSY));
		}
		else
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(), UI_Constants.DECKPOSX, UI_Constants.DECKPOSY));
		}

		if(!gameState.discards.empty())
		{
			final UI_CardPanel currentPanel = new UI_CardPanel(getGameCard(gameState.discards.peek(), GameCardType.PLAYCARD), false);
			final Stack<UI_CardPanel> searchFrameStack = convertToCardPanelStack(gameState.discards, GameCardType.PLAYCARD);
			currentPanel.addMouseListener
			(
				new MouseAdapter()
				{
					public void mouseClicked(MouseEvent e)
					{
						if((e.getButton() == MouseEvent.BUTTON1)&&(e.getClickCount() % 2 == 0))
						{
							new UI_SearchFrame("Discard Pile", searchFrameStack);
						}
					}
				}
			);

			disposableComponents.add(placeMainPanelComponent(currentPanel, UI_Constants.DISCARDSPOSX, UI_Constants.DISCARDSPOSY));
		}
		else
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(), UI_Constants.DISCARDSPOSX, UI_Constants.DISCARDSPOSY));
		}

		renderTimeline(gameState.timeline, gameState.currentInputState, gameState.terminalYear);

		if(gameState.currentInputState == CN_Constants.InputState.GET_LINCHPIN_SELECTION)
		{
			setTimelineVisible(true);
			timelineFrameJustShown = true;
		}
		else if(gameState.currentInputState == CN_Constants.InputState.GET_RIPPLEPOINT_SELECTION)
		{
			setTimelineVisible(true);
			timelineFrameJustShown = true;
		}
		else
		{
			if(timelineFrameJustShown)
				timelineFrameJustShown = false;
			else
				setTimelineVisible(false);
		}

		JLabel unpatchedParadoxLabel = new JLabel("Unpatched Paradoxes: " + String.valueOf(gameState.unpatchedParadoxes));
		unpatchedParadoxLabel.setFont(UI_Constants.UNPATCHED_PARADOXES_FONT);
		if(gameState.unpatchedParadoxes > UI_Constants.UNPATCHED_PARADOX_LABEL_HIGH_THRESHOLD)
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_HIGH_COLOR);
		else if(gameState.unpatchedParadoxes > UI_Constants.UNPATCHED_PARADOX_LABEL_MID_THRESHOLD)
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_MID_COLOR);
		else
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_LOW_COLOR);

		disposableComponents.add(placeMainPanelComponent(unpatchedParadoxLabel, UI_Constants.UNPATCHED_PARADOXES_X, UI_Constants.UNPATCHED_PARADOXES_Y));

		updateStatus("");
		if(gameState.aiMessage != null)
		{
			updateStatus(gameState.aiMessage);
		}
		else if(gameState.currentInputState == CN_Constants.InputState.MAIN)
		{
			updateStatus(playerNames[gameState.currentPlayer] + " has begun.");
		}
		else if(gameState.currentInputState == CN_Constants.InputState.BETWEEN_TURNS)
		{
			updateStatus("It's " + playerNames[gameState.currentPlayer] + "'s turn.");
		}

		validate();
	}



	public void renderGameState(final CN_GameState gameState, boolean computerPlayer) throws UI_FatalException
	{
		renderGameState(gameState, computerPlayer, false);
	}

	public void renderGameState(final CN_GameState gameState, boolean computerPlayer, boolean endGame) throws UI_FatalException
	{
		if(gameState == null) //This occurs if an InputMessageException was caught.
		{
			return;
		}

		if(endGame)
		{
			renderForEndGame(gameState);
			return;
		}
		if(computerPlayer)
		{
			renderForComputer(gameState);
			return;
		}

		timelineFrameJustShown = false;


		if(currentSearchFrame != null)
		{
			currentSearchFrame.dispose();
			currentSearchFrame = null;
		}
		if(currentSelectorFrame != null)
		{
			currentSelectorFrame.dispose();
			currentSelectorFrame = null;
		}

		while(!disposableComponents.isEmpty())
		{
			mainPanel.remove(disposableComponents.get(0));
			disposableComponents.remove(0);
		}

		if(gameState.currentInputState == CN_Constants.InputState.GET_DIRECTION_SELECTION)
		{
			renderDirectionSelectorFrame(getGameCard(gameState.sourceOfMostRecentBascom, GameCardType.PLAYCARD).cardNameFront);
		}
		else if(gameState.currentInputState == CN_Constants.InputState.GET_ID_SELECTION)
		{
			renderIdSelectorFrame(getGameCard(gameState.sourceOfMostRecentBascom, GameCardType.PLAYCARD).cardNameFront, gameState.players);
		}
		else if(gameState.currentInputState == CN_Constants.InputState.GET_MISSION_SELECTION)
		{
			renderMissionSelectorFrame(getGameCard(gameState.sourceOfMostRecentBascom, GameCardType.PLAYCARD).cardNameFront, gameState.players, gameState.currentPlayer);
		}
		else if(gameState.currentInputState == CN_Constants.InputState.GET_DECK_SELECTION)
		{
			renderSearchFrame(getGameCard(gameState.sourceOfMostRecentBascom, GameCardType.PLAYCARD).cardNameFront, gameState.deck);
		}
		else if(gameState.currentInputState == CN_Constants.InputState.GET_DISCARDS_SELECTION)
		{
			renderSearchFrame(getGameCard(gameState.sourceOfMostRecentBascom, GameCardType.PLAYCARD).cardNameFront, gameState.discards);
		}

		updateButtons(gameState.currentInputState, gameState.savedGameStateExists);

		if(gameState.currentInputState != CN_Constants.InputState.BETWEEN_TURNS)
		{
			renderPlayers(gameState.players, gameState.currentPlayer);
		}
		else
		{
			setTimelineVisible(false);
			renderPlayers(gameState.players, UI_Constants.NULLPLAYER);
		}

		if(!gameState.deck.empty())
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(getGameCard(gameState.deck.peek(), GameCardType.PLAYCARD), true), UI_Constants.DECKPOSX, UI_Constants.DECKPOSY));
		}
		else
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(), UI_Constants.DECKPOSX, UI_Constants.DECKPOSY));
		}

		if(!gameState.discards.empty())
		{
			final UI_CardPanel currentPanel = new UI_CardPanel(getGameCard(gameState.discards.peek(), GameCardType.PLAYCARD), false);
			final Stack<UI_CardPanel> searchFrameStack = convertToCardPanelStack(gameState.discards, GameCardType.PLAYCARD);
			currentPanel.addMouseListener
			(
				new MouseAdapter()
				{
					public void mouseClicked(MouseEvent e)
					{
						if((e.getButton() == MouseEvent.BUTTON1)&&(e.getClickCount() % 2 == 0))
						{
							new UI_SearchFrame("Discard Pile", searchFrameStack);
						}
					}
				}
			);

			disposableComponents.add(placeMainPanelComponent(currentPanel, UI_Constants.DISCARDSPOSX, UI_Constants.DISCARDSPOSY));
		}
		else
		{
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(), UI_Constants.DISCARDSPOSX, UI_Constants.DISCARDSPOSY));
		}

		renderTimeline(gameState.timeline, gameState.currentInputState, gameState.terminalYear);

		if(gameState.currentInputState == CN_Constants.InputState.GET_LINCHPIN_SELECTION)
		{
			setTimelineVisible(true);
		}
		else if(gameState.currentInputState == CN_Constants.InputState.GET_RIPPLEPOINT_SELECTION)
		{
			setTimelineVisible(true);
		}

		JLabel unpatchedParadoxLabel = new JLabel("Unpatched Paradoxes: " + String.valueOf(gameState.unpatchedParadoxes));
		unpatchedParadoxLabel.setFont(UI_Constants.UNPATCHED_PARADOXES_FONT);
		if(gameState.unpatchedParadoxes > UI_Constants.UNPATCHED_PARADOX_LABEL_HIGH_THRESHOLD)
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_HIGH_COLOR);
		else if(gameState.unpatchedParadoxes > UI_Constants.UNPATCHED_PARADOX_LABEL_MID_THRESHOLD)
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_MID_COLOR);
		else
			unpatchedParadoxLabel.setForeground(UI_Constants.UNPATCHED_PARADOX_LABEL_LOW_COLOR);

		disposableComponents.add(placeMainPanelComponent(unpatchedParadoxLabel, UI_Constants.UNPATCHED_PARADOXES_X, UI_Constants.UNPATCHED_PARADOXES_Y));

		updateStatus(gameState.currentInputState, gameState.inputStateMessage, gameState.currentPlayer);

		validate();
	}



	private void updateButtons(CN_Constants.InputState inputState, boolean isCancelAllowed)
	{
		toggleTimeline.setEnabled(true);

		ActionListener[] listeners = nextTurn.getActionListeners();
		for(ActionListener listener : listeners)
		{
			nextTurn.removeActionListener(listener);
		}

		if(inputState == CN_Constants.InputState.BETWEEN_TURNS)
		{
			nextTurn.setText("Begin Turn");
			nextTurn.addActionListener
			(
				new ActionListener()
				{
					public void actionPerformed(ActionEvent e)
					{
						messenger.sendBeginTurn();
					}
				}
			);
		}
		else
		{
			nextTurn.setText("End Turn");
			nextTurn.addActionListener
			(
				new ActionListener()
				{
					public void actionPerformed(ActionEvent e)
					{
						messenger.sendFinishTurn();
					}
				}
			);

		}
		if((inputState != CN_Constants.InputState.BETWEEN_TURNS)&&(inputState != CN_Constants.InputState.END_TURN))
		{
			nextTurn.setEnabled(false);
		}
		else
		{
			nextTurn.setEnabled(true);
		}

		if((inputState != CN_Constants.InputState.MAIN)&&(inputState != CN_Constants.InputState.GET_CARD_PLAY))
		{
			discardOne.setEnabled(false);
			discardTwo.setEnabled(false);
		}
		else
		{
			discardOne.setEnabled(true);
			discardTwo.setEnabled(true);
		}

		if((isCancelAllowed)&&(inputState != CN_Constants.InputState.MAIN))
		{
			cancel.setEnabled(true);
		}
		else
		{
			cancel.setEnabled(false);
		}

	}

	private void updateButtonsForComputer()
	{
		toggleTimeline.setEnabled(false);
		nextTurn.setEnabled(false);
		discardOne.setEnabled(false);
		discardTwo.setEnabled(false);
		cancel.setEnabled(false);
	}

	private void updateButtonsForEndGame()
	{
		toggleTimeline.setEnabled(true);
		nextTurn.setEnabled(false);
		discardOne.setEnabled(false);
		discardTwo.setEnabled(false);
		cancel.setEnabled(false);
	}



	private void updateStatus(String status)
	{
		messageLabel.setText(status);
	}

	private void updateStatus(CN_Constants.InputState inputState, String status, int currentPlayer)
	{
		if(inputState == CN_Constants.InputState.MAIN)
			messageLabel.setText("Play a card.");
		else if(inputState == CN_Constants.InputState.BETWEEN_TURNS)
			messageLabel.setText(playerNames[currentPlayer] + ", press \"Begin Turn\" to start your turn.");
		else
			messageLabel.setText(status);
	}


	private void createButtons(CN_Constants.InputState inputState)
	{
		toggleTimeline.setPreferredSize(toggleTimeline.getPreferredSize());
		toggleTimeline.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					if(timeline.isVisible())
					{
						timeline.setVisible(false);
						toggleTimeline.setText("Show Timeline");
					}
					else
					{
						timeline.setVisible(true);
						toggleTimeline.setText("Hide Timeline");
					}
				}
			}
		);
		panelLayout.putConstraint(SpringLayout.WEST, toggleTimeline, 5, SpringLayout.WEST, mainPanel);
		panelLayout.putConstraint(SpringLayout.NORTH, toggleTimeline, 5, SpringLayout.NORTH, mainPanel);
		mainPanel.add(toggleTimeline);


		timeline.addWindowListener
		(
			new WindowAdapter()
			{
				public void windowClosing(WindowEvent e)
				{
					setTimelineVisible(false);
				}
			}
		);


		nextTurn.setPreferredSize(toggleTimeline.getPreferredSize());
		panelLayout.putConstraint(SpringLayout.WEST, nextTurn, 5, SpringLayout.WEST, mainPanel);
		panelLayout.putConstraint(SpringLayout.NORTH, nextTurn, 5, SpringLayout.SOUTH, toggleTimeline);
		mainPanel.add(nextTurn);


		discardOne.setPreferredSize(toggleTimeline.getPreferredSize());
		discardOne.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					messenger.sendDiscardClick();
				}
			}
		);
		panelLayout.putConstraint(SpringLayout.WEST, discardOne, 5, SpringLayout.WEST, mainPanel);
		panelLayout.putConstraint(SpringLayout.NORTH, discardOne, 5, SpringLayout.SOUTH, nextTurn);
		mainPanel.add(discardOne);


		discardTwo.setPreferredSize(toggleTimeline.getPreferredSize());
		discardTwo.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					messenger.sendDiscardTwoClick();
				}
			}
		);
		panelLayout.putConstraint(SpringLayout.WEST, discardTwo, 5, SpringLayout.WEST, mainPanel);
		panelLayout.putConstraint(SpringLayout.NORTH, discardTwo, 5, SpringLayout.SOUTH, discardOne);
		mainPanel.add(discardTwo);


		cancel.setPreferredSize(toggleTimeline.getPreferredSize());
		cancel.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					if(currentSearchFrame != null)
					{
						currentSearchFrame.dispose();
						currentSearchFrame = null;
					}
					if(currentSearchFrame != null)
					{
						currentSearchFrame.dispose();
						currentSearchFrame = null;
					}
					messenger.sendCancel();
				}
			}
		);
		panelLayout.putConstraint(SpringLayout.WEST, cancel, 5, SpringLayout.WEST, mainPanel);
		panelLayout.putConstraint(SpringLayout.NORTH, cancel, 5, SpringLayout.SOUTH, discardTwo);
		mainPanel.add(cancel);

	}





	private void createPlayers(CN_GameState.PlayerData[] currentPlayers) throws UI_FatalException
	{
		if(currentPlayers.length > 4)
		{
			throw new UI_FatalException("UI_MainFrame, createPlayers(): There are more players than the UI can handle!");
		}

		UI_PlayerPane.Orientation orientation;
		UI_PlayerPane newPane;
		JLabel newLabel;
		int xPosition;
		int yPosition;

		for(CN_GameState.PlayerData player : currentPlayers)
		{
			switch(player.playerIndex)
			{
				case 0:
					orientation = UI_PlayerPane.Orientation.SOUTH;
					xPosition = UI_Constants.PANE_POSITION_SOUTH_X;
					yPosition = UI_Constants.PANE_POSITION_SOUTH_Y;
					break;
				case 1:
					orientation = UI_PlayerPane.Orientation.EAST;
					xPosition = UI_Constants.PANE_POSITION_EAST_X;
					yPosition = UI_Constants.PANE_POSITION_EAST_Y;
					break;
				case 2:
					orientation = UI_PlayerPane.Orientation.NORTH;
					xPosition = UI_Constants.PANE_POSITION_NORTH_X;
					yPosition = UI_Constants.PANE_POSITION_NORTH_Y;
					break;
				case 3:
					orientation = UI_PlayerPane.Orientation.WEST;
					xPosition = UI_Constants.PANE_POSITION_WEST_X;
					yPosition = UI_Constants.PANE_POSITION_WEST_Y;
					break;
				default:
					throw new UI_FatalException("UI_MainFrame, createPlayers(): There are more players than the UI can handle, or the player indices do not increment properly!");
			}
			newPane = new UI_PlayerPane(orientation);
			playerPanes[player.playerIndex] = newPane;
			placeMainPanelComponent(newPane, xPosition, yPosition);

			newLabel = new JLabel(playerNames[player.playerIndex]);
			newLabel.setFont(UI_Constants.PLAYER_NAME_FONT);
			nameLabels[player.playerIndex] = newLabel;
			switch(player.playerIndex)
			{
				case 0:
					panelLayout.putConstraint(SpringLayout.WEST, newLabel, 5, SpringLayout.EAST, newPane);
					panelLayout.putConstraint(SpringLayout.SOUTH, newLabel, 0, SpringLayout.SOUTH, newPane);
					break;
				case 1:
					panelLayout.putConstraint(SpringLayout.EAST, newLabel, 0, SpringLayout.EAST, newPane);
					panelLayout.putConstraint(SpringLayout.NORTH, newLabel, 5, SpringLayout.SOUTH, newPane);
					break;
				case 2:
					panelLayout.putConstraint(SpringLayout.EAST, newLabel, -5, SpringLayout.WEST, newPane);
					panelLayout.putConstraint(SpringLayout.NORTH, newLabel, 0, SpringLayout.NORTH, newPane);
					break;
				case 3:
					panelLayout.putConstraint(SpringLayout.WEST, newLabel, 0, SpringLayout.WEST, newPane);
					panelLayout.putConstraint(SpringLayout.NORTH, newLabel, 5, SpringLayout.SOUTH, newPane);
					break;
				default:
					throw new UI_FatalException("UI_MainFrame, createPlayers(): There are more players than the UI can handle, or the player indices do not increment properly!");
			}
			mainPanel.add(newLabel);


			JLabel personalityLabel;
			personalityLabel = new JLabel(personalityNames[player.playerIndex]);
			personalityLabel.setFont(UI_Constants.PLAYER_PERSONALITY_FONT);
			personalityLabel.setForeground(UI_Constants.PLAYER_PERSONALITY_COLOR);
			personalityNameLabels[player.playerIndex] = personalityLabel;
			switch(player.playerIndex)
			{
				case 0:
					panelLayout.putConstraint(SpringLayout.WEST, personalityLabel, 5, SpringLayout.EAST, newPane);
					panelLayout.putConstraint(SpringLayout.SOUTH, personalityLabel, -5, SpringLayout.NORTH, newLabel);
					break;
				case 1:
					panelLayout.putConstraint(SpringLayout.EAST, personalityLabel, 0, SpringLayout.EAST, newPane);
					panelLayout.putConstraint(SpringLayout.NORTH, personalityLabel, 5, SpringLayout.SOUTH, newLabel);
					break;
				case 2:
					panelLayout.putConstraint(SpringLayout.EAST, personalityLabel, -5, SpringLayout.WEST, newPane);
					panelLayout.putConstraint(SpringLayout.NORTH, personalityLabel, 5, SpringLayout.SOUTH, newLabel);
					break;
				case 3:
					panelLayout.putConstraint(SpringLayout.WEST, personalityLabel, 0, SpringLayout.WEST, newPane);
					panelLayout.putConstraint(SpringLayout.NORTH, personalityLabel, 5, SpringLayout.SOUTH, newLabel);
					break;
				default:
					throw new UI_FatalException("UI_MainFrame, createPlayers(): There are more players than the UI can handle, or the player indices do not increment properly!");
			}
			mainPanel.add(personalityLabel);

		}
		setPersonalityNameLabelsVisible(false);
	}




	private void renderTimeline(CN_GameState.TimelineCardData[] timelineData, CN_Constants.InputState inputState, int terminalYear) throws UI_FatalException
	{
		if(timelineData.length > UI_Constants.MAX_PER_ROW * UI_Constants.MAX_PER_COL)
		{
			throw new UI_FatalException("UI_TimelineFrame, renderTimeline: The UI can support only 32 cards in the Timeline!");
		}

		UI_TimelineCardPanel[] timelinePanels = new UI_TimelineCardPanel[timelineData.length];

		CN_GameState.TimelineCardData currentData;
		boolean isXedOut = false;
		boolean isUnstable = false;
		for(int i = 0; i < timelineData.length; i++)
		{
			currentData = timelineData[i];

			if(currentData.year > terminalYear)
			{
				isXedOut = true;
			}

			if(currentData.isUnstable)
				isUnstable = true;
			else
				isUnstable = false;

			if(currentData.patchedBy == CN_Constants.NULL_PATCH)
			{
				final UI_GameCard currentCard = UI_Constants.timelineCardDirectory.get(currentData.cardNumber);
				timelinePanels[i] = new UI_TimelineCardPanel(currentCard, currentData.isFlipped, isXedOut, isUnstable);
				timelinePanels[i].addMouseListener
				(
					new MouseAdapter()
					{
						public void mouseClicked(MouseEvent e)
						{
							if((e.getButton() == MouseEvent.BUTTON1)&&(e.getClickCount() % 2 == 0))
							{
								messenger.sendTimelineCardClick(currentCard);
							}
						}
					}
				);
			}
			else
			{
				final UI_GameCard currentCard = UI_Constants.playCardDirectory.get(currentData.patchedBy);
				timelinePanels[i] = new UI_TimelineCardPanel(currentCard, false, isXedOut, isUnstable);
				timelinePanels[i].addMouseListener
				(
					new MouseAdapter()
					{
						public void mouseClicked(MouseEvent e)
						{
							if((e.getButton() == MouseEvent.BUTTON1)&&(e.getClickCount() % 2 == 0))
							{
								messenger.sendPlayCardClick(currentCard);
							}
						}
					}
				);
			}

			if((inputState == CN_Constants.InputState.GET_LINCHPIN_SELECTION)||(inputState == CN_Constants.InputState.GET_RIPPLEPOINT_SELECTION))
			{
				if(currentData.isLegal)
				{
					timelinePanels[i].setMarked(true);
				}
				else
				{
					timelinePanels[i].setMarked(false);
				}
			}
			else
			{
				timelinePanels[i].setMarked(false);
			}


		}

		timeline.placeTimeline(timelinePanels);

	}


	private void renderPlayers(CN_GameState.PlayerData[] players, int currentPlayerIndex) throws UI_FatalException
	{
		renderPlayers(players, currentPlayerIndex, false);
	}


	private void renderPlayers(CN_GameState.PlayerData[] players, int currentPlayerIndex, boolean endGame) throws UI_FatalException
	{
		ArrayList<UI_GameCard> hand;
		ArrayList<UI_GameCard> artifacts;

		ArrayList<UI_CardPanel> handPanels;
		ArrayList<UI_CardPanel> artifactsPanels;

		boolean isPlayersTurn;


		CN_GameState.PlayerData player;
		for(int i = 0; i < players.length; i++)
		{
			player = players[i];

			hand = convertToGameCardList(player.hand, GameCardType.PLAYCARD);
			artifacts = convertToGameCardList(player.artifacts, GameCardType.PLAYCARD);

			if(endGame)
			{
				isPlayersTurn = true;
			}
			else if(player.playerIndex == currentPlayerIndex)
			{
				isPlayersTurn = true;
			}
			else
			{
				isPlayersTurn = false;
			}

			handPanels = convertToCardPanels(hand, isPlayersTurn);
			artifactsPanels = convertToCardPanels(artifacts, true);

			playerPanes[player.playerIndex].renderPane(handPanels, artifactsPanels);

			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(getGameCard(player.id, GameCardType.IDCARD), !isPlayersTurn), UI_Constants.IDPOSX, UI_Constants.IDMISSIONSTARTINGPOSY + UI_Constants.IDMISSIONINCREMENTY * i));
			disposableComponents.add(placeMainPanelComponent(new UI_CardPanel(getGameCard(player.mission, GameCardType.MISSIONCARD), !isPlayersTurn), UI_Constants.MISSIONPOSX, UI_Constants.IDMISSIONSTARTINGPOSY + UI_Constants.IDMISSIONINCREMENTY * i));
		}
	}




	private void renderIdSelectorFrame(String title, CN_GameState.PlayerData[] playerData) throws UI_FatalException
	{
		ArrayList<JButton> buttons = new ArrayList<JButton>();

		JButton currentButton;
		for(int i = 0; i < playerData.length; i++)
		{
			final CN_GameState.PlayerData currentPlayerData = playerData[i];
			currentButton = new JButton();
			currentButton.setText(playerNames[i]);
			buttons.add(currentButton);
		}

		final UI_SelectorFrame newFrame = new UI_SelectorFrame(title, buttons);

		for(int i = 0; i < playerData.length; i++)
		{
			final CN_GameState.PlayerData currentPlayerData = playerData[i];
			currentButton = buttons.get(i);
			currentButton.addActionListener
			(
				new ActionListener()
				{
					public void actionPerformed(ActionEvent e)
					{
						messenger.sendIdCardSelection(UI_Constants.idCardDirectory.get(currentPlayerData.id));
						newFrame.dispose();
					}
				}
			);
		}
		currentSelectorFrame = newFrame;
	}



	private void renderMissionSelectorFrame(String title, CN_GameState.PlayerData[] playerData, int currentPlayer) throws UI_FatalException
	{
		ArrayList<JButton> buttons = new ArrayList<JButton>();

		JButton currentButton;
		for(int i = 0; i < playerData.length; i++)
		{
			final CN_GameState.PlayerData currentPlayerData = playerData[i];
			currentButton = new JButton();
			if(i == currentPlayer)
				currentButton.setText("(Brand New)");
			else
				currentButton.setText(playerNames[i]);
			buttons.add(currentButton);
		}

		final UI_SelectorFrame newFrame = new UI_SelectorFrame(title, buttons);

		for(int i = 0; i < playerData.length; i++)
		{
			final CN_GameState.PlayerData currentPlayerData = playerData[i];
			currentButton = buttons.get(i);
			currentButton.addActionListener
			(
				new ActionListener()
				{
					public void actionPerformed(ActionEvent e)
					{
						messenger.sendMissionCardSelection(UI_Constants.missionCardDirectory.get(currentPlayerData.mission));
						newFrame.dispose();
					}
				}
			);
		}
		currentSelectorFrame = newFrame;
	}




	private void renderDirectionSelectorFrame(String title) throws UI_FatalException
	{
		ArrayList<JButton> buttons = new ArrayList<JButton>();

		JButton button0 = new JButton();
		JButton button1 = new JButton();
		button0.setText("Left");
		button1.setText("Right");
		buttons.add(button0);
		buttons.add(button1);
		final UI_SelectorFrame newFrame = new UI_SelectorFrame(title, buttons);

		button0.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					messenger.sendDirectionSelection(true);
					newFrame.dispose();
				}
			}
		);

		button1.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					messenger.sendDirectionSelection(false);
					newFrame.dispose();
				}
			}
		);
		currentSelectorFrame = newFrame;
	}




	private void renderSearchFrame(String title, Stack<Integer> deck) throws UI_FatalException
	{
		JButton acceptButton = new JButton();

		final UI_SearchFrame newFrame = new UI_SearchFrame(title, convertToCardPanelStack(deck, GameCardType.PLAYCARD), acceptButton);


		acceptButton.setText("Accept");
		acceptButton.addActionListener
		(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					messenger.sendPlayCardClick(newFrame.getSelectedCard());
				}
			}
		);

		currentSearchFrame = newFrame;

	}



	private JComponent placeMainPanelComponent(JComponent component, int xcoord, int ycoord)
	{
		panelLayout.putConstraint(SpringLayout.WEST, component, xcoord, SpringLayout.WEST, mainPanel);
		panelLayout.putConstraint(SpringLayout.NORTH, component, ycoord, SpringLayout.NORTH, mainPanel);
		mainPanel.add(component);
		return component;
	}

	private ArrayList<UI_CardPanel> convertToCardPanels(ArrayList<UI_GameCard> gameCardList, boolean shouldBeFaceUp)
	{
		ArrayList<UI_CardPanel> returnValue = new ArrayList<UI_CardPanel>();

		UI_CardPanel newPanel;
		for(int i = 0; i < gameCardList.size(); i++)
		{
			final UI_GameCard currentCard = gameCardList.get(i);
			newPanel = new UI_CardPanel(currentCard, !shouldBeFaceUp);

			newPanel.addMouseListener
			(
				new MouseAdapter()
				{
					public void mouseClicked(MouseEvent e)
					{
						if((e.getButton() == MouseEvent.BUTTON1)&&(e.getClickCount() % 2 == 0))
						{
							messenger.sendPlayCardClick(currentCard);
						}
					}
				}
			);

			returnValue.add(newPanel);
		}

		return returnValue;

	}

	private ArrayList<UI_GameCard> convertToGameCardList(ArrayList<Integer> list, GameCardType cardType) throws UI_FatalException
	{
		ArrayList<UI_GameCard> returnValue = new ArrayList<UI_GameCard>();

		for(int i = 0; i < list.size(); i++)
		{
			returnValue.add(getGameCard(list.get(i), cardType));
		}

		return returnValue;
	}


	private Stack<UI_CardPanel> convertToCardPanelStack(Stack<Integer> stack, GameCardType cardType) throws UI_FatalException
	{
		Stack<Integer> copiedStack = new Stack<Integer>();
		Stack<Integer> reverseStack = new Stack<Integer>();

		Stack<UI_CardPanel> returnValue = new Stack<UI_CardPanel>();

		for(int i = 0; i < stack.size(); i++)
		{
			copiedStack.push(stack.get(i));
		}

		while(!copiedStack.empty())
		{
			reverseStack.push(copiedStack.pop());
		}

		while(!reverseStack.empty())
		{
			returnValue.push(new UI_CardPanel(getGameCard(reverseStack.pop(), cardType), false));
		}

		return returnValue;

	}


	private UI_GameCard getGameCard(int cardNumber, GameCardType cardType) throws UI_FatalException
	{
		switch(cardType)
		{
			case PLAYCARD:
{
if(UI_Constants.playCardDirectory.get(cardNumber) == null)
throw new UI_FatalException("It's null: " + cardNumber);
				return UI_Constants.playCardDirectory.get(cardNumber);
}
			case TIMELINECARD:
				return UI_Constants.timelineCardDirectory.get(cardNumber);
			case IDCARD:
				return UI_Constants.idCardDirectory.get(cardNumber);
			case MISSIONCARD:
				return UI_Constants.missionCardDirectory.get(cardNumber);
			default:
				throw new UI_FatalException("UI_MainFrame, getGameCard(): invalid card type!");
		}
	}


	private void setTimelineVisible(boolean setVisible)
	{
		if(setVisible)
		{
			timeline.setVisible(true);
			toggleTimeline.setText("Hide Timeline");
		}
		else
		{
			timeline.setVisible(false);
			toggleTimeline.setText("Show Timeline");
		}
	}


	private void setPersonalityNameLabelsVisible(boolean visible)
	{
		for(JLabel label : personalityNameLabels)
		{
			label.setVisible(visible);
		}
	}


	public void performDisposal()
	{
		disposed = true;
		UI_EndOfUniverseDialog.disposeDialog();
		UI_VictoryDialog.disposeDialog();
		timeline.dispose();
		dispose();
	}

	public boolean getDisposed()
	{
		return disposed;
	}

	public String getPlayerName(int playerIndex)
	{
		return playerNames[playerIndex];
	}

}
